"use strict";
/**
*  @license
* Copyright 2021, JsData. All rights reserved.
*
* This source code is licensed under the MIT license found in the
* LICENSE file in the root directory of this source tree.

* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
* ==========================================================================
*/
Object.defineProperty(exports, "__esModule", { value: true });
exports.MinMaxScaler = void 0;
const utils_1 = require("../utils");
const typesUtils_1 = require("../typesUtils");
const math_1 = require("../math");
const mixins_1 = require("../mixins");
const tf_singleton_1 = require("../tf-singleton");
/**
 * Transform features by scaling each feature to a given range.
 * This estimator scales and translates each feature individually such
 * that it is in the given range on the training set, e.g. between the maximum and minimum value.
 *
 * @example
 * ```js
 * import { MinMaxScaler } from 'scikitjs'
 *
 * const data = [
      [-1, 2],
      [-0.5, 6],
      [0, 10],
      [1, 18]
    ]
    const scaler = new MinMaxScaler()
    const expected = scaler.fitTransform(data)
    // const expected = [
    //  [0, 0],
    //  [0.25, 0.25],
    //  [0.5, 0.5],
    //  [1, 1]
    //]
    ```
 */
class MinMaxScaler extends mixins_1.TransformerMixin {
    featureRange;
    /** The per-feature scale that we see in the dataset. */
    scale;
    min;
    /** The per-feature minimum that we see in the dataset. */
    dataMin;
    /** The per-feature maximum that we see in the dataset. */
    dataMax;
    /** The per-feature range that we see in the dataset. */
    dataRange;
    /** The number of features seen during fit */
    nFeaturesIn;
    /** The number of samples processed by the Estimator. Will be reset on new calls to fit */
    nSamplesSeen;
    /** Names of features seen during fit. Only stores feature names if input is a DataFrame */
    featureNamesIn;
    /** Useful for pipelines and column transformers to have a default name for transforms */
    name = 'MinMaxScaler';
    constructor({ featureRange = [0, 1] } = {}) {
        super();
        this.tf = (0, tf_singleton_1.getBackend)();
        this.featureRange = featureRange;
        this.scale = this.tf.tensor1d([]);
        this.min = this.tf.tensor1d([]);
        this.dataMin = this.tf.tensor1d([]);
        this.dataMax = this.tf.tensor1d([]);
        this.dataRange = this.tf.tensor1d([]);
        this.nFeaturesIn = 0;
        this.nSamplesSeen = 0;
        this.featureNamesIn = [];
    }
    isNumber(value) {
        return typeof value === 'number' && isFinite(value);
    }
    /**
     * Fits a MinMaxScaler to the data
     */
    fit(X) {
        (0, typesUtils_1.assert)((0, typesUtils_1.isScikit2D)(X), 'Data can not be converted to a 2D matrix.');
        (0, typesUtils_1.assert)(this.isNumber(this.featureRange[0]) &&
            this.isNumber(this.featureRange[1]) &&
            this.featureRange[0] < this.featureRange[1], 'featureRange needs to contain exactly two numbers where the first is less than the second');
        const tensorArray = (0, utils_1.convertToNumericTensor2D)(X);
        const max = (0, math_1.tensorMax)(tensorArray, 0, true);
        const min = (0, math_1.tensorMin)(tensorArray, 0, true);
        const range = max.sub(min);
        this.scale = this.tf.div(this.featureRange[1] - this.featureRange[0], (0, math_1.turnZerosToOnes)(range));
        this.min = this.tf.sub(this.featureRange[0], min.mul(this.scale));
        this.dataMin = min;
        this.dataMax = max;
        this.dataRange = range;
        this.nSamplesSeen = tensorArray.shape[0];
        this.nFeaturesIn = tensorArray.shape[1];
        if ((0, typesUtils_1.isDataFrameInterface)(X)) {
            this.featureNamesIn = [...X.columns];
        }
        return this;
    }
    /**
     * Transform the data using the fitted scaler
     * */
    transform(X) {
        (0, typesUtils_1.assert)((0, typesUtils_1.isScikit2D)(X), 'Data can not be converted to a 2D matrix.');
        const tensorArray = (0, utils_1.convertToNumericTensor2D)(X);
        const outputData = tensorArray.mul(this.scale).add(this.min);
        return outputData;
    }
    /**
     * Inverse transform the data using the fitted scaler
     * */
    inverseTransform(X) {
        (0, typesUtils_1.assert)((0, typesUtils_1.isScikit2D)(X), 'Data can not be converted to a 2D matrix.');
        const tensorArray = (0, utils_1.convertToNumericTensor2D)(X);
        const outputData = tensorArray.sub(this.min).div(this.scale);
        return outputData;
    }
}
exports.MinMaxScaler = MinMaxScaler;
